
/*

  A little machine that does the same computations as gnu/sed
  (stream-editor)
  
    Copyright (C) 2011  Alin C Soare < as1789@gmail.com >

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <strings.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "gram.h"

#define MAX_TXT_LENGTH 600
#define MAX_ERR_LENGTH 600

char *expression = "p;{{x}}";
char *E;
char *buffer;
char *tok_start;
#define TOKEN (*buffer)

#define stop_buffer(var) { var = TOKEN; TOKEN=0; }
#define restore_end_buffer(var) { TOKEN = var; }

#define NEXT_CHAR_TOKEN read_token (Token_Char, NULL, NULL);

void
print_token (Sed_Object t)
{
    switch (t.o.type)
    {
    case Token_Number:
        printf ("Token Number: %d\n", t.n.n);
        break;
    case Token_Regexp:
        printf ("Token Regexp \n");
        break;
    case Token_Char:
        printf ("Token Char: %c\n", t.c.c);
        break;
    case Token_String:
        printf ("Token String: %s\n", t.s.s);
        break;
    case Token_Void:
        printf ("Token Void: \n");
        break;
    default:
        printf ("Token error [%d] \n", t.o.type);
    }
}

void
print_command (Sed_atomic_command c)
{
    switch (c.c.command)
    {
    case Search:
        printf ("Command Search: \n") ;
        print_token (c.search.s);
        print_token (c.search.r);
        print_token (c.search.f);
        break;
    case Block:
        printf ("Command Block \n");
        break;
    case Label:
        printf ("Command Label: \n") ;
        print_token (c.label.l);
        break;
    case ConditionalBranch:
        printf ("Command C-branch: \n") ;
        print_token (c.cbranch.l);
        break;
    case NonConditionalBranch:
        printf ("Command NC-branch: \n") ;
        print_token (c.ncbranch.l);
        break;
    default:
        printf ("Command %c \n", c.c.command) ;
    }
}

void
print_atomic_address (Atomic_address a)
{
    switch (a.type)
    {
    case Address_Numeric:
        printf (" %d", a.n.n);
        break;
    case Address_Regexp:
        printf (" R");
        break;
    }
}

void
print_address (Address a)
{
    switch (a.a.type)
    {
    case Address_Single:
        printf ("addr single: ");
        if (a.a.negation)
            printf (" ! ");
        print_atomic_address (a.s.a);
        printf ("\n");
        break;
    case Address_Interval:
        printf ("addr int: ");
        if (a.a.negation)
            printf (" ! ");
        print_atomic_address (a.i.a);
        printf (" , ");
        print_atomic_address (a.i.b);
        printf ("\n");
        break;
    }
}

void print_node (Node *n)
{
    printf ("Node:");
    print_address (n->a);
    print_command (n->c);
    if (n->next_match)
    {
        if (n->next_match != n)
            print_node (n->next_match);
    }
}

Sed_Object
read_token ( Token_Type type, char *separator, char *accept )
{
    Sed_Object result;
    char t;

    if (!TOKEN)
    {
        printf ("error reading: end\n");
        result.o.type = Token_Void;
        return result;
    }

    tok_start = buffer;
    switch (type)
    {
    case Token_Number:
        while (isdigit (TOKEN))
            buffer++;
        stop_buffer (t);
        result.n.n = atoi (tok_start);
        restore_end_buffer (t);
        break;
    case Token_Regexp:
    case Token_String:
        if (!separator) separator="\n\t ";
        while (!index(separator, TOKEN))
        {
            buffer++;
            if ('\\'==TOKEN)
                buffer++;
        }
        //stop_buffer (t);
        t= TOKEN;
        TOKEN=0;
        if (type == Token_Regexp)
        {
            //regex_t *new = (regex_t *) malloc (sizeof (regex_t));
            int err;
            if ( (err = regcomp (&result.r.r, tok_start, 0)) != 0 )
            {
                char err_msg[MAX_ERR_LENGTH];
                regerror (err, &result.r.r, err_msg, MAX_ERR_LENGTH);
                printf ("Error regexp %s: %s.\n", tok_start, err_msg);
                /* FATAL ERROR */
                /* throw an error */
            }
            restore_end_buffer (t);
        }
        else
        {
            //char *new = malloc (buffer - tok_start + 1);
            result.s.s = (char *) calloc (1, buffer - tok_start + 1);
            memcpy ( result.s.s, tok_start, buffer - tok_start );
        }
        restore_end_buffer (t);
        break;
    case Token_Char:
        if (accept)
        {
            if (!index (accept, TOKEN))
            {
                /* throw a fatal error */
            }
        }
        result.c.c = TOKEN;
        buffer++;
        break;
    case Token_Void:
        break;
    }

    result.o.type = type;
    /* jump over spaces */
    /* while ( index ("\n\t ", TOKEN) ) buffer++; */
    print_token (result);
    return result;
}

Node command ();

Node*
command_block ()
{
    Node *first, *last;
    Node n;
    n = command ();
    last = first = malloc (sizeof (Node));
    if (Block == n.c.c.command)
    {
        last = n.next_not_match;
    }
    *first = n;
    while (';' == TOKEN)
    {
        NEXT_CHAR_TOKEN;
        Node new = command ();
        if (Block == last->c.c.command)
        {
            if (last->next_not_match->next_match)
            {
                printf ("should not happen !!!");
            }
            last->next_not_match->next_match = malloc (sizeof (Node));
            *last->next_not_match->next_match = new;
        }
        else
        {
            last->next_match = malloc (sizeof (Node));
            *last->next_match=new;
        }
        if ( Block == last->c.c.command )
        {
            last -> next_not_match = last->next_not_match->next_match;
            last = last -> next_not_match;
        }
        else
        {
            last = last->next_match;
        }
        first->next_not_match = last;
    }
    return first;
}

Atomic_address
atomic_address ()
{
    Atomic_address a;
    if (isdigit (TOKEN))
    {
        Sed_Object N;
        a.type = Address_Numeric;
        N = read_token (Token_Number, NULL, NULL);
        a.n.n = N.n.n;
    }
    else if ('/' == TOKEN)
    {
        Sed_Object R;
        a.type = Address_Regexp;
        NEXT_CHAR_TOKEN;
        R = read_token (Token_Regexp, "/", NULL);
        read_token (Token_Char, NULL, "/");
        a.r.r = R.r.r;
    }
    else
        a.type = Atomic_Address_Void;
    return a;
}

Address
address ()
{
    Address A;
    Atomic_address a, b;
    a.type = b.type = Atomic_Address_Void;
    A.a.negation = 0;
    if ('!' == TOKEN)
    {
        A.a.negation = 1;
        NEXT_CHAR_TOKEN;
    }
    if ((isdigit (TOKEN)) || ('/' == TOKEN))
    {
        a = atomic_address ( );
        if (',' == TOKEN)
        {
            NEXT_CHAR_TOKEN;
            b = atomic_address ( );
        }
    }
    else if (A.a.negation)
    {
        /* throw an error: after negation should be an address */
    }
    if ( Atomic_Address_Void != a.type)
    {
        A.i.a = a;
        if ( Atomic_Address_Void != b.type)
        {
            A.a.type = Address_Interval;
            A.i.b = b;
        }
        else
            A.a.type = Address_Single;
    }
    else
        A.a.type = Address_Void;
    print_address (A);
    return A;
}

Sed_atomic_command
atom ()
{
    Sed_atomic_command X;
    Sed_Object tok;

    Sed_command C = TOKEN;

    switch (TOKEN)
    {
    case 's':
    {
        /* search command */
        Sed_command_search S;
        NEXT_CHAR_TOKEN;
        char separator[2] = {0,0};
        tok = NEXT_CHAR_TOKEN; /* separator */
        *separator=(char) tok.c.c;
        S.s = read_token (Token_Regexp, separator, NULL); /* regexp */
        read_token (Token_Char, separator, NULL);         /* separator */
        S.r = read_token (Token_String, separator, NULL); /* replace with */
        read_token (Token_Char, separator, NULL);         /* separator */
        S.f = read_token (Token_String, ";\n\t ", NULL);  /* flags */
        X.search = S;
    }
    break;
    case ':' :
    {
        /* label def */
        Sed_command_label L;
        L.l = read_token (Token_String, ";\n\t ", NULL);
        X.label = L;
    }
    break;
    case 'b':
    {
        /* branch commands */
        Sed_command_non_cond_branch b;
        NEXT_CHAR_TOKEN;
        b.l = read_token (Token_String, ";\n\t ", NULL);
    }
    break;
    case 't':
    {
        /* branch commands */
        Sed_command_cond_branch t;
        NEXT_CHAR_TOKEN;
        t.l = read_token (Token_String, ";\n\t ", NULL);
    }
    break;
    default:
    {
        if (index ("nNpghHxGdDpPq", TOKEN))
        {
            /* command TOKEN */
            NEXT_CHAR_TOKEN;
        }
        else
        {
            /* command not found. fatal */
        }
    }
    }
    X.c.command = C;
    print_command (X);
    return X;
}

/*

  RECURSIVE DEFINITION OF THE DATA STRUCTURES created by the mutually-
  recursive functions Command and Command_block


OPERATION 1 : Command returns an atomic sed command

               +----------+------------+
          +----| Address  | Command    |
          |    +----+-----+-----+------+
          |         |           |
          |         |           |
          |         |           |
  pointer |         v           v
  to atomic         NULL        NULL
  operation

OPERATION 2 : Command_block links nodes that contain only sed atomic commands like this

               +----------+------------+
               | Address  | Command    |
               +---+------+-------+----+
                   |              |
                   |              |
                   |              +--------+
                   v                       |
               +----------+------------+   |
               | Address  | Command    |   |
               +----+ ----+------+-----+   |
                    |            |         |
                    |            |         |
                    |            v         |
                    .            NULL      |
                    .                      |
                    .                      |
                    .             o--------+
                    .             |
                    .             |
                    v             v
               +----------+------------+
               | Address  | Command    |
               +----+-----+------+-----+
                    |            |
                    |            |
                    |            |
                    v            v
                    NULL         NULL


OPERATION 3: Command returns a block that looks like this


               +----------+------------+
          +--> | Address  | Command    |
          |    +----+-----+------+-----+
          |         |            |
          |         |            +-------------------------------+
          |         +-----------+                                |
          |                     |                                |
  pointer /            ,,,,,,,,,|,,,,,,,,,,,,,,,,,,,,,,,,,,,     |
  to Block             ,        v                          ,     |
                       ,   +----------+------------+       ,     |
                       ,   | Address  | Command    |       ,     |
                       ,   +---+------+-------+----+       ,     |
                       ,       |              .            ,     |
                       ,       |              .            ,     |
                       ,       |              ..........   ,     |
                       ,       v                       .   ,     |
                       ,   +----------+------------+   .   ,     |
                       ,   | Address  | Command    |   .   ,     |
                       ,   +----+ ----+------+-----+   .   ,     |
                       ,        |            |         .   ,     |
                       ,        |            |         .   ,     |
                       ,        |            v         .   ,     |
                       ,        .            NULL      .   ,     |
                       ,        .                      .   ,     |
                       ,        .                      .   ,     |
                       ,        .             ..........   ,     |
                       ,        .             . +----------------+
                       ,        .             . |          ,
                       ,        v             v v          ,
                       ,   +----------+------------+       ,
                       ,   | Address  | Command    |       ,
                       ,   +----+-----+------+-----+       ,
                       ,        |            |             ,
                       ,        |            |             ,
                       ,        |            |             ,
                       ,        v            v             ,
                       ,        NULL         NULL          ,
                       ,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,



OPERATION 4 : Command_Block links nodes that are Blocks like this:
                ( asterisk's line replaces old links )

                +----------+------------+
       +------> | Address  | Command    |
       |        +----+-----+------+-----+
       |             |            |
       |             |            |
     LAST            |            |
     is not a block  |            |
                     |            |
                     v            v
                +----------+------------+
           +--> | Address  | Block      |
           |    +----+-----+------+---*-+
           |         |            .   ***********************************
           |         |            .................................     *
           |         +-----------+                                .     *
           |                     |                                .     *
         LAST           ,,,,,,,,,|,,,,,,,,,,,,,,,,,,,,,,,,,,,     .     *
         is a block     ,        v                          ,     .     *
                        ,   +----------+------------+       ,     .     *
                        ,   | Address  | Command    |       ,     .     *
                        ,   +---+------+-------+----+       ,     .     *
                        ,       |              |            ,     .     *
                        ,       |              |            ,     .     *
                        ,       |              +--------+   ,     .     *
                        ,       v                       |   ,     .     *
                        ,   +----------+------------+   |   ,     .     *
                        ,   | Address  | Command    |   |   ,     .     *
                        ,   +----+ ----+------+-----+   |   ,     .     *
                        ,        |            |         |   ,     .     *
                        ,        |            |         |   ,     .     *
                        ,        |            v         |   ,     .     *
                        ,        .            NULL      |   ,     .     *
                        ,        .                      |   ,     .     *
                        ,        .                      |   ,     .     *
                        ,        .             +--------+   ,     .     *
                        ,        .             |...................     *
                        ,        .             | .          ,           *
                        ,        v             v v          ,           *
                        ,   +----------+------------+       ,           *
                        ,   | Address  | Command    |       ,           *
                        ,   +-*--+-----+------+-----+       ,           *
                        ,     *  |            |             ,           *
                        ,     *  |            |             ,           *
                        ,     *  |            |             ,           *
                        ,     *  v            v             ,           *
                        ,     *  NULL         NULL          ,           *
                        ,,,,,,*,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,           *
                              *                                         *
                              *    **************************************
                       ********    *
                       *           *
                       *           *
                       v           v
                +----------+------------+
     NEW -----> | Address  | Command    |
                +----+-----+------+-----+

*/

Node
command ()
{
    Node new;
    new.a=address ();
    if ('{' == TOKEN)
    {
        NEXT_CHAR_TOKEN;
        Node *n = command_block ();
        new . c . c. command = Block;
        new . next_match = n;
        new . next_not_match = n -> next_not_match;
        read_token (Token_Char, NULL, "}");
    }
    else
    {
        new.c = atom ();
        /* on ne peut pas savoir a cette place les liaisons */
        new. next_match = new. next_not_match = NULL;
    }
    return new;
}

void
evaluate_sed_expression (Node *expr)
{
    Node *e;
    e = expr;
    while (e)
    {
        if (ConditionalBranch == e->c.c.command)
        {
            /* t is a special form */
            /* use lazy-evaluation */
        }
        if (NonConditionalBranch == e->c.c.command)
        {
            /* b is a special form */
            /* use lazy-evaluation */
            /* for every node X from `first`, if X->next_match is the
             * current e, then look for the node Y labeled with the
             * label of the current branch, and link X->next_match =
             * Y->next_match. This memoization will make the current
             * node E never to be reached , and can be freed.
             */
        }
        {
            /* the rest are functions and are called via apply */
            /* Apply e->c.c.action (); */
        }
    }
}

main ()
{
    E=malloc (strlen (expression)+1);
    strcpy (E, expression);
    buffer = E;
    Node *n = command_block ();
    printf ("------------\n");
    print_node (n);
}

