
/* COMMAND */
typedef
enum
{
    Token_Number,
    Token_String,
    Token_Regexp,
    Token_Char,
    Token_Void
} Token_Type;

typedef
struct
{
    Token_Type type;
    int n;
} Sed_Number;

typedef
struct
{
    Token_Type type;
    char *s;
} Sed_String;

typedef
struct
{
    Token_Type type;
    regex_t r;
} Sed_Regexp;

typedef
struct
{
    Token_Type type;
    char c;
} Sed_Char;

typedef
struct
{
    Token_Type type;
    void *data;
} Sed_Void;

typedef
union
{
    struct 
    {
        Token_Type type;
        void *data;
    } o;
    Sed_Number n;
    Sed_String s;
    Sed_Regexp r;
    Sed_Char   c;
    Sed_Void   v;
} Sed_Object;

typedef
enum
{
    Search = 's',
    Label = ':',
    ConditionalBranch = 't',
    NonConditionalBranch = 'b',
    Nextn = 'n',
    NextN ='N',
    Getg ='g',
    GetG ='G',
    Holdh = 'h',
    HoldH ='H',
    Exchange ='x',
    Deleted = 'd',
    DeleteD = 'D',
    Printp = 'p',
    PrintP = 'P',
    Quit = 'q',
    Block = -1
} Sed_command;

typedef
struct
{
    Sed_command command;
    void *action;
    Sed_Object s;             /* search */
    Sed_Object r;             /* replace */
    Sed_Object f;             /* flags */
} Sed_command_search;

typedef
struct
{
    Sed_command command;
    void *action;
    Sed_Object l;
} Sed_command_label;

typedef
struct
{
    Sed_command command;
    void *action;
    Sed_Object l;
} Sed_command_cond_branch;

typedef
struct
{
    Sed_command command;
    void *action;
    Sed_Object l;
} Sed_command_non_cond_branch;

typedef
union
{
    struct
    {
        Sed_command command;
        void *action;
    } c;
    Sed_command_search search;
    Sed_command_label label;
    Sed_command_cond_branch cbranch;
    Sed_command_non_cond_branch ncbranch;
} Sed_atomic_command;

/* ADDRESS */
typedef
enum
{
    Address_Numeric,
    Address_Regexp,
    Atomic_Address_Void
} Atomic_Adress_type;

typedef
enum
{
    Address_Single,
    Address_Interval,
    Address_Void,
} Adress_type;

typedef
struct
{
    Atomic_Adress_type type;
    int n;
} Numeric_Address;

typedef
struct
{
    Atomic_Adress_type type;
    regex_t r;
} Regexp_Address;

typedef
union
{
    Atomic_Adress_type type;
    Numeric_Address n;
    Regexp_Address r;
} Atomic_address;

typedef
struct
{
    Adress_type type;
    char negation;
    Atomic_address a;
} Address_single;

typedef
struct
{
    Adress_type type;
    char negation;
    Atomic_address a;
    Atomic_address b;
} Address_interval;

typedef
union
{
    struct
    {
        Adress_type type;
        char negation;
    } a;
    Address_single s;
    Address_interval i;
} Address;

/* GRAPH */

typedef
struct node
{
    Address a;
    Sed_atomic_command c;
    struct node *next_match;
    struct node *next_not_match;
} Node;
